"""
Pulling Lists of KEGG Entry IDs
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Functionality for pulling lists of KEGG entry IDs from the KEGG REST API.
"""
import typing as t

from . import rest as r


class EntryIdsGetter:
    """
    Class with methods for various ways of pulling a list of KEGG entry IDs from the KEGG REST API.
    """

    def __init__(self, kegg_rest: r.KEGGrest = None) -> None:
        """
        :param kegg_rest: Optional KEGGrest object for making requests to the KEGG REST API to pull entry IDs (if a KEGGrest object is not provided, a KEGGrest object is created with the default values).
        """
        self._kegg_rest = kegg_rest if kegg_rest is not None else r.KEGGrest()

    def from_database(self, database_name: str) -> list:
        """ Pulls the KEGG entry IDs of a given database.

        :param database_name: The KEGG database to pull the entry IDs from.
        :return: The list of resulting entry IDs.
        :raises RuntimeError: Raised if the request to the KEGG REST API fails or times out.
        """
        kegg_response: r.KEGGresponse = self._kegg_rest.list(database_name=database_name)

        return EntryIdsGetter._process_response(kegg_response=kegg_response)


    @staticmethod
    def _process_response(kegg_response: r.KEGGresponse) -> list:
        """ Extracts the entry IDs from a KEGG response if successful, else raises an exception.

        :param kegg_response: The response from the KEGG REST API containing entry IDs if successful.
        :return: The list of KEGG entry IDs.
        :raises RuntimeError: Raised if the KEGG response indicates a failure or time out.
        """
        if kegg_response.status == r.KEGGresponse.Status.FAILED:
            raise RuntimeError(
                f'The KEGG request failed to pull the entry IDs from the following URL: {kegg_response.kegg_url.url}'
            )
        elif kegg_response.status == r.KEGGresponse.Status.TIMEOUT:
            raise RuntimeError(
                f'The KEGG request timed out while trying to pull the entry IDs from the following URL: '
                f'{kegg_response.kegg_url.url}'
            )

        entry_ids: list = EntryIdsGetter._parse_entry_ids_string(entry_ids_string=kegg_response.text_body)

        return entry_ids


    @staticmethod
    def from_file(file_path: str) -> list:
        """ Loads KEGG entry IDs that are listed in a file with one entry ID on each line.

        :param file_path: The path to the file containing the entry IDs.
        :return: The list of entry IDs.
        :raises ValueError: Raised if the file is empty.
        """
        with open(file_path, 'r') as file:
            entry_ids: str = file.read()

            if entry_ids == '':
                raise ValueError(f'Attempted to load entry IDs from {file_path}. But the file is empty')

            entry_ids: list = EntryIdsGetter._parse_entry_ids_string(entry_ids_string=entry_ids)

        return entry_ids


    @staticmethod
    def _parse_entry_ids_string(entry_ids_string: str) -> list:
        """ Parses the entry IDs contained in a string.

        :param entry_ids_string: The string containing the entry IDs.
        :return: The list of parsed entry IDs.
        """
        entry_ids: list = entry_ids_string.strip().split('\n')
        entry_ids = [entry_id.split('\t')[0].strip() for entry_id in entry_ids]

        return entry_ids


    def from_keywords(self, database_name: str, keywords: list) -> list:
        """ Pulls entry IDs from a database based on keywords searched in the entries.

        :param database_name: The name of the database to pull entry IDs from.
        :param keywords: The keywords to search entries in the database with.
        :return: The list of entry IDs.
        :raises RuntimeError: Raised if the request to the KEGG REST API fails or times out.
        """
        kegg_response: r.KEGGresponse = self._kegg_rest.keywords_find(database_name=database_name, keywords=keywords)

        return EntryIdsGetter._process_response(kegg_response=kegg_response)


    def from_molecular_attribute(
        self, database_name: str, formula: str = None, exact_mass: t.Union[float, tuple] = None,
        molecular_weight: t.Union[int, tuple] = None
    ) -> list:
        """ Pulls entry IDs from a database containing chemical entries based on one (and only one) of three molecular attributes of the entries.

        :param database_name: The name of the database containing chemical entries.
        :param formula: The chemical formula to search for.
        :param exact_mass: The exact mass of the compound to search for (a single value or a range).
        :param molecular_weight: The molecular weight of the compound to search for (a single value or a range).
        :return: The list of entry IDs.
        :raises RuntimeError: Raised if the request to the KEGG REST API fails or times out.
        """
        kegg_response: r.KEGGresponse = self._kegg_rest.molecular_find(
            database_name=database_name, formula=formula, exact_mass=exact_mass, molecular_weight=molecular_weight
        )

        return EntryIdsGetter._process_response(kegg_response=kegg_response)



"""
Usage:
    kegg_pull entry-ids -h | --help
    kegg_pull entry-ids from-database <database-name> [--output=<output>] [--zip-file=<zip-file>]
    kegg_pull entry-ids from-file <file-path> [--output=<output>] [--zip-file=<zip-file>]
    kegg_pull entry-ids from-keywords <database-name> <keywords> [--output=<output>] [--zip-file=<zip-file>]
    kegg_pull entry-ids from-molecular-attribute <database-name> (--formula=<formula>|--exact-mass=<exact-mass>...|--molecular-weight=<molecular-weight>...) [--output=<output>] [--zip-file=<zip-file>]

Options:
    -h --help                               Show this help message.
    from-database                           Pulls all the entry IDs within a given database.
    <database-name>                         The KEGG database from which to pull a list of entry IDs.
    --output=<output>                       Path to the file to store the output (1 entry ID per line). Prints to the console if not specified. If ends in ".zip", saves file to a zip archive.
    --zip-file=<zip-file>                   The name of the file to store in a zip archive. If not set, defaults to saving a file with the same name as the zip archive minus the .zip extension. Ignored if --output does not end in ".zip".
    from-file                               Loads the entry IDs from a file.
    <file-path>                             Path to a file containing a list of entry IDs with one entry ID on each line.
    from-keywords                           Searches for entries within a database based on provided keywords.
    <keywords>                              Comma separated list of keywords to search within entries (e.g. --keywords=kw1,k2w,kw3 etc.).
    from-molecular-attribute                Searches a database of molecule-type KEGG entries by molecular attributes.
    --formula=<formula>                     Sequence of atoms in a chemical formula format to search for (e.g. "O5C7" searchers for molecule entries containing 5 oxygen atoms and/or 7 carbon atoms).
    --exact-mass=<exact-mass>               Either a single number (e.g. --exact-mass=155.5) or two numbers (e.g. --exact-mass=155.5 --exact-mass=244.4). If a single number, searches for molecule entries with an exact mass equal to that value rounded by the last decimal point. If two numbers, searches for molecule entries with an exact mass within the two values (a range).
    --molecular-weight=<molecular-weight>   Same as --exact-mass but searches based on the molecular weight.
"""
import docopt as d
import zipfile as zf

from . import entry_ids as ei
from . import _utils as u


def main():
    args: dict = d.docopt(__doc__)
    database_name: str = args['<database-name>']
    entry_ids_getter = ei.EntryIdsGetter()

    if args['from-database']:
        entry_ids: list = entry_ids_getter.from_database(database_name=database_name)
    elif args['from-file']:
        entry_ids_file_path: str = args['<file-path>']
        entry_ids: list = ei.EntryIdsGetter.from_file(file_path=entry_ids_file_path)
    elif args['from-keywords']:
        keywords: str = args['<keywords>']
        keywords: list = u.split_comma_separated_list(list_string=keywords)
        entry_ids: list = entry_ids_getter.from_keywords(database_name=database_name, keywords=keywords)
    else:
        formula, exact_mass, molecular_weight = u.get_molecular_attribute_args(args=args)

        entry_ids: list = entry_ids_getter.from_molecular_attribute(
            database_name=database_name, formula=formula, exact_mass=exact_mass, molecular_weight=molecular_weight
        )

    output: str = args['--output']
    entry_ids: str = '\n'.join(entry_ids)

    if output is not None:
        if output.endswith('.zip'):
            u.save_to_zip_archive(zip_archive_path=output, zip_file_name=args['--zip-file'], file_content=entry_ids)
        else:
            with open(output, 'w') as file:
                file.write(entry_ids)
    else:
        print(entry_ids)



"""
Usage:
    kegg_pull -h | --help       Show this help message.
    kegg_pull -v | --version    Displays the package version.
    kegg_pull --full-help       Show the help message of all sub commands.
    kegg_pull entry-ids ...     Obtain a list of KEGG entry IDs.
    kegg_pull rest ...          Executes one of the KEGG REST API operations.
    kegg_pull pull ...          Pull, separate, and store KEGG entries to the local file system.
"""
import sys

from . import __version__
from . import entry_ids_cli as ei_cli
from . import rest_cli as r_cli
from . import pull_cli as p_cli


def main():
    if len(sys.argv) > 1 and sys.argv[1] == 'entry-ids':
        ei_cli.main()
    elif len(sys.argv) > 1 and sys.argv[1] == 'rest':
        r_cli.main()
    elif len(sys.argv) > 1 and sys.argv[1] == 'pull':
        p_cli.main()
    elif len(sys.argv) > 1 and sys.argv[1] == '--full-help':
        separator = '-'*80
        print(__doc__)
        print(separator)
        print(ei_cli.__doc__)
        print(separator)
        print(r_cli.__doc__)
        print(separator)
        print(p_cli.__doc__)
    elif len(sys.argv) > 1 and (sys.argv[1] == '--version' or sys.argv[1] == '-v'):
        print(__version__)
    else:
        print(__doc__)


if __name__ == '__main__':  # pragma: no cover
    main()  # pragma: no cover



"""
KEGG REST API Operations
~~~~~~~~~~~~~~~~~~~~~~~~
Interface for the KEGG REST API including all its operations.
"""
import enum as e
import requests as rq
import time as t
import inspect as ins
import logging as l

from . import kegg_url as ku


class KEGGresponse:
    """Class containing details of a response from the KEGG REST API."""

    class Status(e.Enum):
        """The status of a KEGG response."""
        SUCCESS = 1
        FAILED = 2
        TIMEOUT = 3

    def __init__(self, status: Status, kegg_url: ku.AbstractKEGGurl, text_body: str = None, binary_body: bytes = None) -> None:
        """
        :param status: The status of the KEGG response.
        :param kegg_url: The URL used in the request to the KEGG REST API that resulted in the KEGG response.
        :param text_body: The text version of the response body.
        :param binary_body: The binary version of the response body.
        :raises ValueError: Raised if the status is SUCCESS but a response body is not provided.
        """
        if status == KEGGresponse.Status.SUCCESS and (
            text_body is None or binary_body is None or text_body == '' or binary_body == b''
        ):
            raise ValueError('A KEGG response cannot be marked as successful if its response body is empty')

        self._status = status
        self._kegg_url = kegg_url
        self._text_body = text_body
        self._binary_body = binary_body

    @property
    def status(self) -> Status:
        return self._status

    @property
    def kegg_url(self) -> ku.AbstractKEGGurl:
        return self._kegg_url

    @property
    def text_body(self) -> str:
        return self._text_body

    @property
    def binary_body(self) -> bytes:
        return self._binary_body


class KEGGrest:
    """Class containing methods for making requests to the KEGG REST API, including all the KEGG REST API operations."""

    def __init__(self, n_tries: int = 3, time_out: int = 60, sleep_time: float = 5.0):
        """
        :param n_tries: The number of times to try to make a request (can succeed the first time, or any of n_tries, or none of the tries).
        :param time_out: The number of seconds to wait for a request until marking it as timed out.
        :param sleep_time: The number of seconds to wait in between timed out requests or blacklisted requests.
        """
        self._n_tries = n_tries if n_tries is not None else 3
        self._time_out = time_out if time_out is not None else 60
        self._sleep_time = sleep_time if sleep_time is not None else 5.0

        if self._n_tries < 1:
            raise ValueError(f'{self._n_tries} is not a valid number of tries to make a KEGG request.')

    def request(self, KEGGurl: type = None, kegg_url: ku.AbstractKEGGurl = None, **kwargs) -> KEGGresponse:
        """ General KEGG request function based on a given KEGG URL (either a class that is instantiated or an already instantiated KEGG URL object).

        :param KEGGurl: Optional KEGG URL class (extended from AbstractKEGGurl) that's instantiated with provided keyword arguments.
        :param kegg_url: Optional KEGGurl object that's already instantiated (used if KEGGurl class is not provided).
        :param kwargs: The keyword arguments used to instantiate the KEGGurl class, if provided.
        :return: The KEGG response.
        """
        kegg_url: ku.AbstractKEGGurl = KEGGrest._get_kegg_url(KEGGurl=KEGGurl, kegg_url=kegg_url, **kwargs)
        status = None

        for _ in range(self._n_tries):
            try:
                response: rq.Response = rq.get(url=kegg_url.url, timeout=self._time_out)

                if response.status_code == 200:
                    return KEGGresponse(
                        status=KEGGresponse.Status.SUCCESS, kegg_url=kegg_url, text_body=response.text,
                        binary_body=response.content
                    )
                else:
                    status = KEGGresponse.Status.FAILED

                if response.status_code == 403:
                    # 403 forbidden. KEGG may have blocked the request due to too many requests in too little time.
                    # In case blacklisting, sleep to allow time for KEGG to unblock further requests.
                    t.sleep(self._sleep_time)
            except rq.exceptions.Timeout:
                status = KEGGresponse.Status.TIMEOUT
                t.sleep(self._sleep_time)

        return KEGGresponse(status=status, kegg_url=kegg_url)

    @staticmethod
    def _get_kegg_url(KEGGurl: type = None, kegg_url: ku.AbstractKEGGurl = None, **kwargs) -> ku.AbstractKEGGurl:
        """ Gets the KEGGurl object to be used to make the request to KEGG.

        :param KEGGurl: Optional KEGGurl class to instantiate a KEGGurl object using keyword arguments.
        :param kegg_url: Instantiated KEGGurl object that's simply returned if provided (used if the KEGGurl class is not provided).
        :param kwargs: The keyword arguments used to instantiate the KEGGurl object if a KEGGurl class is provided.
        :return: The KEGGurl object.
        :raises ValueError: Raised if both a class and object are provided or the class does not inherit from AbstractKEGGurl.
        """
        if KEGGurl is None and kegg_url is None:
            raise ValueError(
                f'Either an instantiated kegg_url object must be provided or an extended class of '
                f'{ku.AbstractKEGGurl.__name__} along with the corresponding kwargs for its constructor.'
            )

        if kegg_url is not None and KEGGurl is not None:
            l.warning(
                'Both an instantiated kegg_url object and KEGGurl class are provided. Using the instantiated object...'
            )

        if kegg_url is not None:
            return kegg_url

        if ku.AbstractKEGGurl not in ins.getmro(KEGGurl):
            raise ValueError(
                f'The value for KEGGurl must be an inherited class of {ku.AbstractKEGGurl.__name__}. '
                f'The class "{KEGGurl.__name__}" is not.'
            )

        kegg_url: ku.AbstractKEGGurl = KEGGurl(**kwargs)

        return kegg_url

    def test(self, KEGGurl: type = None, kegg_url: ku.AbstractKEGGurl = None, **kwargs) -> bool:
        """ Tests if a KEGGurl will succeed upon being used in a request to the KEGG REST API.

        :param KEGGurl: Optional KEGGurl class used to instantiate a KEGGurl object given keyword arguments.
        :param kegg_url: KEGGurl object that's already instantiated (used if a KEGGurl class is not provided).
        :param kwargs: The keyword arguments used to instantiated the KEGGurl object from the KEGGurl class, if provided.
        :return: True if the URL would succeed, false if it would fail or time out.
        """
        kegg_url: ku.AbstractKEGGurl = KEGGrest._get_kegg_url(KEGGurl=KEGGurl, kegg_url=kegg_url, **kwargs)

        for _ in range(self._n_tries):
            try:
                response: rq.Response = rq.head(url=kegg_url.url, timeout=self._time_out)

                if response.status_code == 200:
                    return True
            except rq.exceptions.Timeout:
                t.sleep(self._sleep_time)

        return False

    def list(self, database_name: str) -> KEGGresponse:
        """ Executes the "list" KEGG API operation, pulling the entry IDs of the provided database.

        :param database_name: The database from which to pull entry IDs.
        :return: The KEGG response.
        """
        return self.request(KEGGurl=ku.ListKEGGurl, database_name=database_name)

    def get(self, entry_ids: list, entry_field: str = None) -> KEGGresponse:
        """ Executes the "get" KEGG API operation, pulling the entries of the provided entry IDs.

        :param entry_ids: The IDs of entries to pull.
        :param entry_field: Optional field to extract from the entries.
        :return: The KEGG response.
        """
        return self.request(KEGGurl=ku.GetKEGGurl, entry_ids=entry_ids, entry_field=entry_field)

    def info(self, database_name: str) -> KEGGresponse:
        """ Executes the "info" KEGG API operation, pulling information about a KEGG database.

        :param database_name: The database to pull information about.
        :return: The KEGG response
        """
        return self.request(KEGGurl=ku.InfoKEGGurl, database_name=database_name)

    def keywords_find(self, database_name: str, keywords: list) -> KEGGresponse:
        """ Executes the "find" KEGG API operation, finding entry IDs based on keywords to search in entries.

        :param database_name: The name of the database containing entries to search for.
        :param keywords: The keywords to search in entries.
        :return: The KEGG response
        """
        return self.request(KEGGurl=ku.KeywordsFindKEGGurl, database_name=database_name, keywords=keywords)

    def molecular_find(
        self, database_name: str, formula: str = None, exact_mass: float = None, molecular_weight: int = None
    ) -> KEGGresponse:
        """ Executes the "find" KEGG API operation, finding entry IDs in chemical databases based on one (and only one) choice of three molecular attributes of the entries.

        :param database_name: The name of the chemical database to search for entries in.
        :param formula: The chemical formula (one of three choices) of chemical entries to search for.
        :param exact_mass: The exact mass (one of three choices) of chemical entries to search for (single value or range).
        :param molecular_weight: The molecular weight (one of three choices) of chemical entries to search for (single value or range).
        :return: The KEGG response
        """
        return self.request(
            KEGGurl=ku.MolecularFindKEGGurl, database_name=database_name, formula=formula, exact_mass=exact_mass,
            molecular_weight=molecular_weight
        )

    def database_conv(self, kegg_database_name: str, outside_database_name: str) -> KEGGresponse:
        """ Executes the "conv" KEGG API operation, converting the entry IDs of a KEGG database to those of an outside database.

        :param kegg_database_name: The name of the KEGG database to pull converted entry IDs from.
        :param outside_database_name: The name of the outside database to pull converted entry IDs from.
        :return: The KEGG response.
        """
        return self.request(
            KEGGurl=ku.DatabaseConvKEGGurl, kegg_database_name=kegg_database_name,
            outside_database_name=outside_database_name
        )

    def entries_conv(self, target_database_name: str, entry_ids: list) -> KEGGresponse:
        """ Executes the "conv" KEGG API operation, converting provided entry IDs from one database to the form of a target database.

        :param target_database_name: The name of the database to get converted entry IDs from.
        :param entry_ids: The entry IDs to convert to the form of the target database.
        :return: The KEGG response.
        """
        return self.request(
            KEGGurl=ku.EntriesConvKEGGurl, target_database_name=target_database_name, entry_ids=entry_ids
        )

    def database_link(self, target_database_name: str, source_database_name: str) -> KEGGresponse:
        """ Executes the "link" KEGG API operation, showing the IDs of entries in one KEGG database that are connected/related to entries of another KEGG database.

        :param target_database_name: One of the two KEGG databases to pull linked entries from.
        :param source_database_name: The other KEGG database to link entries from the target database.
        :return: The KEGG response
        """
        return self.request(
            KEGGurl=ku.DatabaseLinkKEGGurl, target_database_name=target_database_name,
            source_database_name=source_database_name
        )

    def entries_link(self, target_database_name: str, entry_ids: list) -> KEGGresponse:
        """ Executes the "link" KEGG API operation, showing the IDs of entries that are connected/related to entries of a provided databases.

        :param target_database_name: The KEGG database to find links to the provided entries.
        :param entry_ids: The IDs of the entries to link to entries in the target database.
        :return: The KEGG response
        """
        return self.request(
            KEGGurl=ku.EntriesLinkKEGGurl, target_database_name=target_database_name, entry_ids=entry_ids
        )

    def ddi(self, drug_entry_ids: list) -> KEGGresponse:
        """ Executes the "ddi" KEGG API operation, searching for drug to drug interactions. Providing one entry ID reports all known interactions, while providing multiple checks if any drug pair in a given set of drugs is CI or P. If providing multiple, all entries must belong to the same database.

        :param drug_entry_ids: The IDs of the drug entries within which search for drug interactions.
        :return: The KEGG response
        """
        return self.request(KEGGurl=ku.DdiKEGGurl, drug_entry_ids=drug_entry_ids)
